/*
 * Common routines.
 *
 * Copyright (C) 2019, 2020, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <libgen.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "templates.h"



/* Generate an error message for file opening failure. */

void open_error(const char *dir, const char *name)
{
  fprintf(stderr, "Could not open source file for writing in directory %s (name %s).\n", dir, name);
}

/* Obtain a new basename string. */

char *make_basename(const char *path)
{
  char s[strlen(path) + 1];

  strcpy(s, path);
  return strdup(basename(s));
}

/* Obtain a new directory name string. */

char *make_dirname(const char *path)
{
  char s[strlen(path) + 1];

  strcpy(s, path);
  return strdup(dirname(s));
}

/* Obtain a file pointer using a parameterised filename. */

FILE *get_output_file(const char *filename_template, const char *dir, const char *name)
{
  FILE *fp;
  char *filename;

  if (asprintf(&filename, filename_template, dir, name) == -1)
    fp = NULL;
  else
  {
    fp = fopen(filename, "w");
    free(filename);
  }

  /* Write an error message upon failure. */

  if (fp == NULL)
    open_error(dir, name);

  return fp;
}

/* Find an attribute value. */

char *get_attribute_value(struct attribute *attr, const char *name)
{
  if (attr == NULL)
    return NULL;

  if (!strcmp(attr->attribute, name) && (attr->identifiers != NULL))
    return attr->identifiers->identifier;

  return get_attribute_value(attr->tail, name);
}

/* Return any conversion to a different object type in a dispatcher. */

char *get_object_conversion(struct interface *iface, int compound)
{
  char *conversion;

  if (!compound)
    return strdup("_self");

  if (asprintf(&conversion, "convert_to_%s(_self)", iface->name) == -1)
    conversion = NULL;

  return conversion;
}

/* Return any opcode attribute value. */

char *get_opcode(struct attribute *attr)
{
  return get_attribute_value(attr, "opcode");
}

/* Return any protocol attribute value. */

char *get_protocol(struct attribute *attr)
{
  return get_attribute_value(attr, "protocol");
}

/* Return the wrapper function prefix. */

char *get_operation_wrapper_prefix(struct attribute *attr)
{
  if (have_attribute(attr, "completion"))
    return "initiate";
  else
    return "wrap";
}

int have_attribute(struct attribute *attr, const char *name)
{
  if (attr == NULL)
    return 0;

  if (!strcmp(attr->attribute, name))
    return 1;

  return have_attribute(attr->tail, name);
}

int have_attribute_value(struct attribute *attr, const char *name,
                         const char *value)
{
  if (attr == NULL)
    return 0;

  if (!strcmp(attr->attribute, name) && have_identifier(attr->identifiers, value))
    return 1;

  return have_attribute_value(attr->tail, name, value);
}

int have_identifier(struct identifier *ident, const char *value)
{
  if (ident == NULL)
    return 0;

  if (!strcmp(ident->identifier, value))
    return 1;

  return have_identifier(ident->tail, value);
}

/* Return the function role of the given signature for server use. */

enum function_role get_server_function_role(struct signature *sig)
{
  return have_attribute(sig->attributes, "completion") ? COMPLETION_ROLE
                                                       : GENERAL_FUNCTION_ROLE;
}

/* Obtain a suitable interface class name depending on the component role. */

char *get_interface_class_name(struct interface *iface,
                               enum component_role component)
{
  char *name;

  switch (component)
  {
    case CLIENT_ROLE:
      if (asprintf(&name, "client_%s", iface->name) == -1)
        name = NULL;
      break;

    case SERVER_ROLE:
      name = strdup(iface->name);
      break;

    default:
      name = NULL;
      break;
  }

  return name;
}

/* Generate a qualified operation name for declarations and definitions. */

char *get_signature_operation_name(struct interface *iface,
                                   struct signature *sig,
                                   enum component_role component,
                                   enum signature_role role, enum language lang)
{
  char *opname, *class_name;

  switch (lang)
  {
    /* In C, interface members use the bare names. */

    case C_LANGUAGE:
      if ((role == MEMBER_DECLARATION_ROLE) || (role == ACCESS_ROLE))
        opname = strdup(sig->operation);
      else
        opname = get_operation_name(iface, sig);
      break;

    /* In C++, method definitions are qualified using class names. */

    case CPP_LANGUAGE:
      if (role == DEFINITION_ROLE)
      {
        class_name = get_interface_class_name(iface, component);

        if (asprintf(&opname, "%s::%s", class_name, sig->operation) == -1)
          opname = NULL;

        free(class_name);
      }
      else
        opname = strdup(sig->operation);
      break;

    default:
      opname = NULL;
      break;
  }

  return opname;
}

/* Generate a qualified operation name for structure and function naming. */

char *get_operation_name(struct interface *iface, struct signature *sig)
{
  char *opname;

  if (asprintf(&opname, "%s_%s", iface->name, sig->operation) == -1)
    opname = NULL;

  return opname;
}

/* Return the invocation opcode given protocol and operation name details. */

char *get_opcode_identifier(char *protocol, char *opname)
{
  char *opcode;

  if (protocol != NULL)
    return protocol;

  if (asprintf(&opcode, "opcode_%s", opname) == -1)
    return NULL;

  return opcode;
}

/* Return the appropriate parameter decoration. Generally, output parameters
   involve addresses. However, completion functions accept output parameters as
   values. */

const char *get_parameter_decoration(struct parameter *param,
                                     enum parameter_role role,
                                     enum function_role function)
{
  if ((param->specifier & OUT_PARAMETER) && (function != COMPLETION_ROLE))
  {
    switch (role)
    {
      case SIGNATURE_ROLE: return "*";
      case INVOCATION_ROLE: return "&";
      default: break;
    }
  }

  return "";
}

/* Return the parameter name, being the final identifier in a list. */

char *get_parameter_name(struct identifier *ident)
{
  if (ident->tail != NULL)
    return get_parameter_name(ident->tail);
  else
    return ident->identifier;
}

/* Get the appropriate terminating text after a signature. */

char *get_signature_terminator(enum signature_role role)
{
  /* Signatures in declarations or member initialisation. */

  if (role & DECLARATION_ROLE)
    return COMPLETE_SIGNATURE;

  /* Signatures at the start of functions. */

  if (role & DEFINITION_ROLE)
    return BEGIN_FUNCTION;

  return "";
}

/* Count the different kinds of parameters in terms of inputs, outputs, words
   and items. */

void count_parameters(struct parameter *param, int *input_words,
                      int *input_items, int *output_words, int *output_items)
{
  for (*input_words = 0, *input_items = 0, *output_words = 0, *output_items = 0;
       param != NULL;
       param = param->tail)
  {
    /* Special case: "out fpage". */

    if ((param->specifier & IN_PARAMETER) || (param->cls == FPAGE_ITEM))
    {
      if (param->cls & ITEM_CLASS) (*input_items)++;
      else (*input_words)++;
    }

    if (param->specifier & OUT_PARAMETER)
    {
      if (param->cls & ITEM_CLASS) (*output_items)++;
      else (*output_words)++;
    }
  }
}

/* Return the maximum number of input items accepted by any one of the given
   signatures. */

int get_max_input_items(struct signature *sig)
{
  int input_words, input_items, output_words, output_items;
  int max_input_items = 0;

  for (; sig != NULL; sig = sig->tail)
  {
    count_parameters(sig->parameters, &input_words, &input_items, &output_words, &output_items);

    if (input_items > max_input_items)
      max_input_items = input_items;
  }

  return max_input_items;
}

/* Return any operator needed to access the given parameter as a local name. */

char *access_name(struct parameter *param, enum component_role component,
                  int writing)
{
  /* In the client, employ the dereferencing operator for "inout" and "out"
     word parameters since the parameters will employ pointer types. */

  if ((component == CLIENT_ROLE) && (param->specifier & OUT_PARAMETER) &&
      (param->cls & WORD_CLASS))
    return "*";

  /* In the client, employ the dereferencing operator for written "inout"
     item parameters since the parameters will employ pointer types. */

  if ((component == CLIENT_ROLE) && (param->specifier & OUT_PARAMETER) &&
      (param->cls & ITEM_CLASS) && writing)
    return "*";

  /* In the server, when importing items from the message, the address of the
     variable to hold the item must be used. */

  if ((component == SERVER_ROLE) && (param->specifier & IN_PARAMETER) &&
      (param->cls & ITEM_CLASS) && !writing)
    return "&";

  return "";
}

/* Return the item type name for the given parameter. */

char *item_type_name(struct parameter *param, enum component_role component,
                     int writing)
{
  if (param->cls == FPAGE_ITEM)
  {
    /* Permit the specification of receive window flexpage items. */

    if (specify_receive_fpage(param, component, writing))
      return "receive_fpage";
    else
      return "fpage";
  }
  else
    return "capability";
}

/* Return whether the parameter involves a receive window flexpage. */

int specify_receive_fpage(struct parameter *param,
                          enum component_role component, int writing)
{
  return (param->cls == FPAGE_ITEM) &&
         (component == CLIENT_ROLE) &&
         !(param->specifier & IN_PARAMETER) &&
         (param->specifier & OUT_PARAMETER) &&
         writing;
}

/* When copying output items from the message (in the client) or copying into
   the message (in the server, but only with a new message), the message address
   must be obtained using the appropriate operator. */

char *reference_message(enum component_role component,
                        enum function_role function)
{
  return (component == CLIENT_ROLE) || (function == COMPLETION_ROLE) ? "&" : "";
}

/* Return the naming prefix for structure access to a message. */

char *structure_prefix(enum specifier direction)
{
  return direction == IN_PARAMETER ? "in" : "out";
}

/* Return whether message access for the given component role and specifier (or
   direction) involves writing to (or populating) the message. */

int writing_to_message(enum component_role component, enum specifier direction)
{
  return (component == CLIENT_ROLE) && (direction & IN_PARAMETER) ||
         (component == SERVER_ROLE) && (direction & OUT_PARAMETER);
}
